
from itertools import combinations
import sys

SIZE_UPPER_LIMIT = 6

# s inputs
# t ouputs
# r gates

def input_clause(n,i,input):
	neg_input = list(set(range(1,n+1))-set(input))
	return "I%d := AND("%i + ', '.join(map(lambda j:  "x%d_%d" % (i,j), input)) + ", " + ', '.join(map(lambda x:  "~x%d_%d" % (i,x), neg_input)) + ");"

def comp_clause(n,i,j,k,disjoint = False):
	bits = map(lambda s: "(x%d_%d | x%d_%d) == x%d_%d" % (i,s,j,s,k,s), range(1,n+1))
	disjoints = []
	if disjoint:
		disjoints = map(lambda s: "~x%d_%d | ~x%d_%d" % (i,s,j,s), range(1,n+1))
	return "C%d_%d_%d := AND(" % (k,i,j) + ', '.join(bits + disjoints) + ");"

def comp_clauses(n,k,disjoint = False):
	clauses = []
	for j in range(2, k):
		for i in range(1,j):
			clauses.append(comp_clause(n,i,j,k, disjoint))
	l = []
	for j in range(2, k):
		for i in range(1,j):
			l.append("C%d_%d_%d" % (k,i,j))
	clauses.append("C%d := OR(" % k + ', '.join(l) + ");")
	return clauses

def output_clause(n,j,output):
	neg_output = list(set(range(1,n+1))-set(output))
	return "AND(" + ', '.join(map(lambda k:  "x%d_%d" % (j,k), output) + map(lambda k:  "~x%d_%d" % (j,k), neg_output)) + ")"

def output_clauses(n,i, min, max,output):
	return "O%d := OR(" % i + ', '.join(map(lambda j: output_clause(n,j,output), range(min, max+1))) + ");"

def lex_order_clause(n,i,j):
	# j larger than i
	return "L%d_%d := AND((~x%d_1 | x%d_1), " % (i,j,i,j) + ', '.join(map( lambda k: 'AND(' + ', '.join(map(lambda l: 'x%d_%d == x%d_%d' % (i,l,j,l), range(1,k+1))) + ') => (~x%d_%d | x%d_%d)' % (i, k+1, j, k+1) ,range(1, n)))    + ');' 

def size_clause(n, i, k):
	return 'S%d := [1,%d](' % (i, k) + ', '.join(['x%d_%d' % (i,j) for j in range(1, n+1)]) + ');'

def formula(n, r, outputs, antichain = False, disjoint = False):
	frm = []
	frm.append("BC1.0")
	s = n
	inputs = [[i] for i in range(1,n+1)]
	t = len(outputs)
	max_size_limit = min(SIZE_UPPER_LIMIT, n-2, r-s-t)
	for i in range(s):
		frm.append(input_clause(n,i+1,inputs[i]))
	for i in range(1, max_size_limit+1):
		frm.append(size_clause(n, s+i, i+1))
	for k in range(n+1, r+1):
		frm.extend(comp_clauses(n,k, disjoint))
	if antichain:
		for i in range(t):
			frm.append(output_clauses(n, i+1 , r-t+i+1, r-t+i+1, outputs[i]))
		for i in range(s+1,r-t):
			for j in range(i+1, r-t+1):
				frm.append(lex_order_clause(n, i, j))
	else:
		for i in range(t):
			frm.append(output_clauses(n, i+1 , s+1, r, outputs[i]))
		for i in range(s+1,r):
			for j in range(i+1, r+1):
				frm.append(lex_order_clause(n, i, j))
	head = "f := AND(" + ', '.join(["I%d" % i for i in range(1, s+1)]) + ', ' + \
		    	  ', '.join(["O%d" % i for i in range(1, t+1)]) + ', ' + \
		    	  ', '.join(["C%d" % i for i in range(s+1, r+1)])
	lex = ''
	if antichain:
		if s + 1 < r-t:
			lex = ', ' + ', '.join(['L%d_%d' % (i,j) for i in range(s+1,r-t) for j in range(i+1,r-t+1)])
	else:
		lex = ', ' + ', '.join(['L%d_%d' % (i,j) for i in range(s+1,r) for j in range(i+1,r+1)])
	size = ''
	if max_size_limit > 1:
		  size = ', ' + ', '.join(["S%d" % i for i in range(s+1, s+max_size_limit+1)])
	frm.append(head + lex + size + ');')
	frm.append("ASSIGN f;")
	return '\n'.join(frm)
